@@ -11,11 +11,13 @@ enum {
 
 void vgcs_rach(struct gsm_lchan *lchan, uint8_t ra, uint8_t acc_delay, uint32_t fn);
 
+void vgcs_lchan_activate(struct gsm_lchan *lchan);
+
 void vgcs_lchan_react(struct gsm_lchan *lchan);
 
 void vgcs_talker_frame(struct gsm_lchan *lchan);
 
-void vgcs_talker_reset(struct gsm_lchan *lchan);
+void vgcs_talker_reset(struct gsm_lchan *lchan, bool ul_access);
 
 void vgcs_listener_reset(struct gsm_lchan *lchan);
 
