@@ -13,7 +13,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -33,6 +33,8 @@
 #include <osmocom/gsm/l1sap.h>
 #include <osmocom/gsm/gsm_utils.h>
 #include <osmocom/gsm/rsl.h>
+#include <osmocom/gsm/rlp.h>
+#include <osmocom/gsm/rtp_extensions.h>
 #include <osmocom/core/gsmtap.h>
 #include <osmocom/core/gsmtap_util.h>
 #include <osmocom/core/utils.h>
@@ -957,10 +959,10 @@ static void l1sap_update_fnstats(struct gsm_bts *bts, uint32_t rts_fn)
 }
 
 /* Common dequeueing function */
-static inline struct msgb *lapdm_phsap_dequeue_msg(struct lapdm_entity *le)
+static inline struct msgb *lapdm_phsap_dequeue_msg(struct lapdm_entity *le, uint32_t fn)
 {
 	struct osmo_phsap_prim pp;
-	if (lapdm_phsap_dequeue_prim(le, &pp) < 0)
+	if (lapdm_phsap_dequeue_prim_fn(le, &pp, fn) < 0)
 		return NULL;
 	return pp.oph.msg;
 }
@@ -983,7 +985,7 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_facch(struct gsm_lchan *lchan
 		lchan->rep_acch.dl_facch[1].msg = NULL;
 	} else {
 		/* Fetch new FACCH from queue ... */
-		if (lapdm_phsap_dequeue_prim(le, &pp) < 0)
+		if (lapdm_phsap_dequeue_prim_fn(le, &pp, fn) < 0)
 			return NULL;
 		msg = pp.oph.msg;
 
@@ -1013,7 +1015,7 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_facch(struct gsm_lchan *lchan
 }
 
 /* Special dequeueing function with SACCH repetition (3GPP TS 44.006, section 11) */
-static inline struct msgb *lapdm_phsap_dequeue_msg_sacch(struct gsm_lchan *lchan, struct lapdm_entity *le)
+static inline struct msgb *lapdm_phsap_dequeue_msg_sacch(struct gsm_lchan *lchan, struct lapdm_entity *le, uint32_t fn)
 {
 	struct osmo_phsap_prim pp;
 	struct msgb *msg;
@@ -1037,7 +1039,7 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_sacch(struct gsm_lchan *lchan
 	}
 
 	/* Fetch new repetition candidate from queue */
-	if (lapdm_phsap_dequeue_prim(le, &pp) < 0)
+	if (lapdm_phsap_dequeue_prim_fn(le, &pp, fn) < 0)
 		return NULL;
 	msg = pp.oph.msg;
 	sapi = (msg->data[0] >> 2) & 0x07;
@@ -1148,9 +1150,9 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 						LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-SACCH repetition: active => inactive\n");
 					lchan->rep_acch.dl_sacch_active = false;
 				}
-				pp_msg = lapdm_phsap_dequeue_msg_sacch(lchan, le);
+				pp_msg = lapdm_phsap_dequeue_msg_sacch(lchan, le, fn);
 			} else {
-				pp_msg = lapdm_phsap_dequeue_msg(le);
+				pp_msg = lapdm_phsap_dequeue_msg(le, fn);
 			}
 		} else {
 			if (lchan->ts->trx->bts->dtxd)
@@ -1159,7 +1161,7 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 			if (lchan->rep_acch.dl_facch_active && lchan->rsl_cmode != RSL_CMOD_SPD_SIGN)
 				pp_msg = lapdm_phsap_dequeue_msg_facch(lchan, le, fn);
 			else
-				pp_msg = lapdm_phsap_dequeue_msg(le);
+				pp_msg = lapdm_phsap_dequeue_msg(le, fn);
 			lchan->tch.dtx_fr_hr_efr.dl_facch_stealing = (pp_msg != NULL);
 		}
 		if (!pp_msg) {
@@ -1802,9 +1804,18 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 		lchan_meas_handle_sacch(lchan, msg);
 	}
 
+	if (L1SAP_IS_LINK_SACCH(link_id))
+		le = &lchan->lapdm_ch.lapdm_acch;
+	else
+		le = &lchan->lapdm_ch.lapdm_dcch;
+
 	/* bad frame */
-	if (len == 0)
+	if (len == 0) {
+		/* Notify current receive FN to lapdm. */
+		lapdm_t200_fn(le, fn);
+
 		return -EINVAL;
+	}
 
 	/* report first valid received frame to handover process */
 	if (lchan->ho.active == HANDOVER_WAIT_FRAME)
@@ -1825,11 +1836,6 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 			return 0;
 	}
 
-	if (L1SAP_IS_LINK_SACCH(link_id))
-		le = &lchan->lapdm_ch.lapdm_acch;
-	else
-		le = &lchan->lapdm_ch.lapdm_dcch;
-
 	if (check_for_first_ciphrd(lchan, data, len))
 		l1sap_tx_ciph_req(lchan->ts->trx, chan_nr, 1, 0);
 
@@ -1837,23 +1843,111 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 	msgb_pull_to_l2(msg);
 	lapdm_phsap_up(&l1sap->oph, le);
 
+	/* Notify current receive FN to lapdm. */
+	lapdm_t200_fn(le, fn);
+
 	/* don't free, because we forwarded data */
 	return 1;
 }
 
-static void send_ul_rtp_packet_data(struct gsm_lchan *lchan, uint32_t fn,
+/* process one MAC block of unpacked bits of a non-transparent CSD channel */
+static void gsmtap_csd_rlp_process(struct gsm_lchan *lchan, bool is_uplink,
+				   const struct ph_tch_param *tch_ind,
+				   const uint8_t *data, unsigned int data_len)
+{
+	struct gsm_bts_trx *trx = lchan->ts->trx;
+	struct gsmtap_inst *inst = trx->bts->gsmtap.inst;
+	struct osmo_rlp_frame_decoded rlpf;
+	pbit_t *rlp_buf;
+	uint16_t arfcn;
+	int byte_len;
+
+	if (!inst || !trx->bts->gsmtap.rlp)
+		return;
+
+	if (lchan->csd_mode != LCHAN_CSD_M_NT)
+		return;
+
+	if (is_uplink)
+		rlp_buf = lchan->tch.csd.rlp_buf_ul;
+	else
+		rlp_buf = lchan->tch.csd.rlp_buf_dl;
+
+	/* TCH/F 9.6: 4x60bit block => 240bit RLP frame
+	 * TCH/F 4.8: 2x 2x60bit blocks starting at B0/B2/B4 => 240bit RLP frame
+	 * TCH/H 4.8: 4x60bit block => 240bit RLP frame
+	 * TCH/F 2.4: 2x36bit blocks => transparent only
+	 * TCH/H 2.4: 4x36bit blocks => transparent only
+	 * TCH/F 14.4: 2x 290 bit block (starting with M1=0) => 576-bit RLP frame
+	 */
+
+	if (lchan->type == GSM_LCHAN_TCH_F && lchan->tch_mode == GSM48_CMODE_DATA_6k0) {
+		/* in this mode we have 120bit MAC blocks; two of them need to be concatenated
+		 * to render a 240-bit RLP frame. The fist block is present in B0/B2/B4.
+		 * The E7 bit is used to indicate the Frame MF0a */
+		OSMO_ASSERT(data_len == 120);
+		ubit_t e7 = data[4*7+3];
+		if (e7 == 0) {
+			osmo_ubit2pbit_ext(rlp_buf, 0, data, 0, data_len, 1);
+			return;
+		} else {
+			osmo_ubit2pbit_ext(rlp_buf, 120, data, 0, data_len, 1);
+			byte_len = 240/8;
+		}
+	} else if (lchan->type == GSM_LCHAN_TCH_F && lchan->tch_mode == GSM48_CMODE_DATA_14k5) {
+		/* in this mode we have 290bit MAC blocks containing M1, M2 and 288 data bits;
+		 * two of them need to be concatenated to render a
+		 * 576-bit RLP frame. The start of a RLP frame is
+		 * denoted by a block with M1-bit set to 0. */
+		OSMO_ASSERT(data_len == 290);
+		ubit_t m1 = data[0];
+		if (m1 == 0) {
+			osmo_ubit2pbit_ext(rlp_buf, 0, data, 2, data_len, 1);
+			return;
+		} else {
+			osmo_ubit2pbit_ext(rlp_buf, 288, data, 2, data_len, 1);
+			byte_len = 576/8;
+		}
+	} else {
+		byte_len = osmo_ubit2pbit_ext(rlp_buf, 0, data, 0, data_len, 1);
+	}
+
+	if (trx->bts->gsmtap.rlp_skip_null) {
+		int rc = osmo_rlp_decode(&rlpf, 0, rlp_buf, byte_len);
+		if (rc == 0 && rlpf.ftype == OSMO_RLP_FT_U && rlpf.u_ftype == OSMO_RLP_U_FT_NULL)
+			return;
+	}
+
+	arfcn = trx->arfcn;
+	if (is_uplink)
+		arfcn |= GSMTAP_ARFCN_F_UPLINK;
+
+	gsmtap_send_ex(inst, GSMTAP_TYPE_GSM_RLP, arfcn, lchan->ts->nr,
+		       lchan->type == GSM_LCHAN_TCH_H ? GSMTAP_CHANNEL_VOICE_H : GSMTAP_CHANNEL_VOICE_F,
+		       lchan->nr, tch_ind->fn, tch_ind->rssi, 0, rlp_buf, byte_len);
+
+}
+
+static void send_ul_rtp_packet_data(struct gsm_lchan *lchan, const struct ph_tch_param *tch_ind,
 				    const uint8_t *data, uint16_t data_len)
 {
+	struct gsm_bts *bts = lchan->ts->trx->bts;
 	uint8_t rtp_pl[RFC4040_RTP_PLEN];
 	int rc;
 
+	gsmtap_csd_rlp_process(lchan, true, tch_ind, data, data_len);
+
 	rc = csd_v110_rtp_encode(lchan, &rtp_pl[0], data, data_len);
 	if (rc < 0)
 		return;
 
+	rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_TX_TOTAL);
+	if (lchan->rtp_tx_marker)
+		rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_TX_MARKER);
+
 	osmo_rtp_send_frame_ext(lchan->abis_ip.rtp_socket,
 				&rtp_pl[0], sizeof(rtp_pl),
-				fn_ms_adj(fn, lchan),
+				fn_ms_adj(tch_ind->fn, lchan),
 				lchan->rtp_tx_marker);
 	/* Only clear the marker bit once we have sent a RTP packet with it */
 	lchan->rtp_tx_marker = false;
@@ -1863,12 +1957,17 @@ static void send_ul_rtp_packet_data(struct gsm_lchan *lchan, uint32_t fn,
 static void send_ul_rtp_packet_speech(struct gsm_lchan *lchan, uint32_t fn,
 				      const uint8_t *rtp_pl, uint16_t rtp_pl_len)
 {
+	struct gsm_bts *bts = lchan->ts->trx->bts;
+
 	if (lchan->abis_ip.osmux.use) {
 		lchan_osmux_send_frame(lchan, rtp_pl, rtp_pl_len,
 				       fn_ms_adj(fn, lchan), lchan->rtp_tx_marker);
 	} else if (lchan->abis_ip.rtp_socket) {
 		osmo_rtp_send_frame_ext(lchan->abis_ip.rtp_socket,
 			rtp_pl, rtp_pl_len, fn_ms_adj(fn, lchan), lchan->rtp_tx_marker);
+		rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_TX_TOTAL);
+		if (lchan->rtp_tx_marker)
+			rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_TX_MARKER);
 	}
 	/* Only clear the marker bit once we have sent a RTP packet with it */
 	lchan->rtp_tx_marker = false;
@@ -1892,13 +1991,42 @@ static void send_rtp_rfc5993(struct gsm_lchan *lchan, uint32_t fn,
 	send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
 }
 
+/* a helper function for emitting FR/EFR UL in TW-TS-001 format */
+static void send_rtp_twts001(struct gsm_lchan *lchan, uint32_t fn,
+			     struct msgb *msg, bool good_frame)
+{
+	uint8_t teh;
+	bool send_frame;
+
+	if (msg->len == GSM_FR_BYTES || msg->len == GSM_EFR_BYTES) {
+		if (good_frame)
+			teh = 0xE0;
+		else
+			teh = 0xE2;
+		send_frame = true;
+	} else {
+		teh = 0xE6;
+		send_frame = false;
+	}
+	/* always set DTXd and TAF bits */
+	if (lchan->ts->trx->bts->dtxd)
+		teh |= 0x08;
+	if (fn % 104 == 52)
+		teh |= 0x01;
+	if (send_frame) {
+		msgb_push_u8(msg, teh);
+		send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
+	} else {
+		send_ul_rtp_packet_speech(lchan, fn, &teh, 1);
+	}
+}
+
 /* A helper function for l1sap_tch_ind(): handling BFI
  *
- * Please note that we pass the msgb to this function, even though it is
- * currently not used.  This msgb passing is a provision for adding
- * support for TRAU-UL-like RTP payload formats like TW-TS-001 that allow
- * indicating BFI along with deemed-bad frame data bits, just like
- * GSM 08.60 and 08.61 TRAU-UL frames.
+ * Please note that the msgb passed to this function is used only when
+ * the CN asked the BSS to emit extended RTP formats (currently TW-TS-001,
+ * later TW-TS-002 as well) that can indicate BFI along with deemed-bad
+ * frame data bits, just like GSM 08.60 and 08.61 TRAU-UL frames.
  */
 static void tch_ul_bfi_handler(struct gsm_lchan *lchan,
 			       const struct gsm_time *g_time, struct msgb *msg)
@@ -1907,6 +2035,20 @@ static void tch_ul_bfi_handler(struct gsm_lchan *lchan,
 	uint8_t ecu_out[GSM_FR_BYTES];
 	int rc;
 
+	/* Are we on TCH/FS or TCH/EFS, configured to emit TW-TS-001 extended
+	 * RTP format?  If so, emit BFI per that spec.  The placement of
+	 * this check before the ECU is intentional: the modes of TW-TS-001
+	 * UL output (closely replicating the classic GSM architecture in which
+	 * a BTS never applies an ECU to its UL output) and internal UL ECU
+	 * are mutually exclusive. */
+	if ((lchan->abis_ip.rtp_extensions & OSMO_RTP_EXT_TWTS001) &&
+	    lchan->type == GSM_LCHAN_TCH_F &&
+	    (lchan->tch_mode == GSM48_CMODE_SPEECH_V1 ||
+	     lchan->tch_mode == GSM48_CMODE_SPEECH_EFR)) {
+		send_rtp_twts001(lchan, fn, msg, false);
+		return;
+	}
+
 	/* Are we applying an ECU to this uplink, and are we in a state
 	 * (not DTX pause) where we emit ECU output? */
 	if (lchan->ecu_state && !osmo_ecu_is_dtx_pause(lchan->ecu_state)) {
@@ -1961,6 +2103,12 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 		LOGPLCGT(lchan, &g_time, DL1P, LOGL_DEBUG, "Rx TCH.ind\n");
 	}
 
+	/* Notify current receive FN to lapdm.
+	 * TCH frames may be indicated before FACCH frames are indicated. To prevent T200 timeout before FACCH is
+	 * received, subtract one frame number, so that timeout is processed next time after FACCH is received.
+	 */
+	lapdm_t200_fn(&lchan->lapdm_ch.lapdm_dcch, GSM_TDMA_FN_SUB(fn, 1));
+
 	/* The ph_tch_param contained in the l1sap primitive may contain
 	 * measurement data. If this data is present, forward it for
 	 * processing */
@@ -1980,14 +2128,30 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 		/* hand msg to RTP code for transmission */
 		switch (lchan->rsl_cmode) {
 		case RSL_CMOD_SPD_SPEECH:
-			if (bts->emit_hr_rfc5993 && lchan->type == GSM_LCHAN_TCH_H &&
-			    lchan->tch_mode == GSM48_CMODE_SPEECH_V1)
-				send_rtp_rfc5993(lchan, fn, msg);
-			else
+			/* support different RTP output formats per codec */
+			if (lchan->type == GSM_LCHAN_TCH_F &&
+			    (lchan->tch_mode == GSM48_CMODE_SPEECH_V1 ||
+			     lchan->tch_mode == GSM48_CMODE_SPEECH_EFR)) {
+				/* FR and EFR codecs */
+				if (lchan->abis_ip.rtp_extensions & OSMO_RTP_EXT_TWTS001)
+					send_rtp_twts001(lchan, fn, msg, true);
+				else
+					send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
+			} else if (lchan->type == GSM_LCHAN_TCH_H &&
+				   lchan->tch_mode == GSM48_CMODE_SPEECH_V1) {
+				/* HR codec: TS 101 318 or RFC 5993,
+				 * will also support TW-TS-002 in the future. */
+				if (bts->emit_hr_rfc5993)
+					send_rtp_rfc5993(lchan, fn, msg);
+				else
+					send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
+			} else {
+				/* generic case, no RTP alterations */
 				send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
+			}
 			break;
 		case RSL_CMOD_SPD_DATA:
-			send_ul_rtp_packet_data(lchan, fn, msg->data, msg->len);
+			send_ul_rtp_packet_data(lchan, tch_ind, msg->data, msg->len);
 			break;
 		case RSL_CMOD_SPD_SIGN:
 			return 0; /* drop stale TCH.ind */
@@ -2266,17 +2430,25 @@ void l1sap_rtp_rx_cb(struct osmo_rtp_socket *rs, const uint8_t *rtp_pl,
 		     uint32_t timestamp, bool marker)
 {
 	struct gsm_lchan *lchan = rs->priv;
+	struct gsm_bts *bts = lchan->ts->trx->bts;
 	struct msgb *msg;
 	bool rfc5993_sid = false;
 
+	rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_RX_TOTAL);
+	if (marker)
+		rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_RX_MARKER);
+
 	/* if we're in loopback mode, we don't accept frames from the
 	 * RTP socket anymore */
-	if (lchan->loopback)
+	if (lchan->loopback) {
+		rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_RX_DROP_LOOPBACK);
 		return;
+	}
 
 	/* initial preen */
 	switch (rtp_payload_input_preen(lchan, rtp_pl, rtp_pl_len, &rfc5993_sid)) {
 	case PL_DECISION_DROP:
+		rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_RX_DROP_PREEN);
 		return;
 	case PL_DECISION_ACCEPT:
 		break;
@@ -2296,8 +2468,13 @@ void l1sap_rtp_rx_cb(struct osmo_rtp_socket *rs, const uint8_t *rtp_pl,
 		int rc = csd_v110_rtp_decode(lchan, msg->tail,
 					     rtp_pl, rtp_pl_len);
 		if (rc > 0) {
+			/* 'fake' tch_ind containing all-zero so gsmtap code can be shared
+			 * between UL and DL */
+			static const struct ph_tch_param fake_tch_ind = {};
+			gsmtap_csd_rlp_process(lchan, false, &fake_tch_ind, msg->tail, rc);
 			msgb_put(msg, rc);
 		} else {
+			rate_ctr_inc2(bts->ctrs, BTS_CTR_RTP_RX_DROP_V110_DEC);
 			msgb_free(msg);
 			return;
 		}
@@ -2407,6 +2584,10 @@ int l1sap_chan_rel(struct gsm_bts_trx *trx, uint8_t chan_nr)
 		lchan->ecu_state = NULL;
 	}
 
+	/* reset CSD RLP buffers to avoid any user plane data leaking from
+	 * one previous lchan into a later one */
+	memset(&lchan->tch.csd, 0, sizeof(lchan->tch.csd));
+
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_DEACTIVATE,
 		0);
 }
@@ -2444,3 +2625,26 @@ int l1sap_chan_modify(struct gsm_bts_trx *trx, uint8_t chan_nr)
 
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_MODIFY, 0);
 }
+
+int l1sap_uplink_access(struct gsm_lchan *lchan, bool active)
+{
+	uint8_t chan_nr = gsm_lchan2chan_nr(lchan);
+	struct osmo_phsap_prim l1sap;
+
+	if (lchan->state != LCHAN_S_ACTIVE && !rsl_chan_rt_is_asci(lchan->rsl_chan_rt)) {
+		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Channel %s is not an active ASCI type channel.\n",
+			  rsl_chan_nr_str(chan_nr));
+		return -EINVAL;
+	}
+
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "%s uplink access detection on channel %s\n",
+		  (active) ? "Activating" : "Deactivating", rsl_chan_nr_str(chan_nr));
+
+
+	memset(&l1sap, 0, sizeof(l1sap));
+	osmo_prim_init(&l1sap.oph, SAP_GSM_PH, PRIM_MPH_INFO, PRIM_OP_REQUEST, NULL);
+	l1sap.u.info.type = (active) ? PRIM_INFO_ACT_UL_ACC : PRIM_INFO_DEACT_UL_ACC;
+	l1sap.u.info.u.ulacc_req.chan_nr = chan_nr;
+
+	return l1sap_down(lchan->ts->trx, &l1sap);
+}
