@@ -13,7 +13,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
@@ -164,9 +164,9 @@ int oml_mo_send_msg(const struct gsm_abis_mo *mo, struct msgb *msg, uint8_t msg_
 	return oml_send_msg(msg, 0);
 }
 
-static inline void add_bts_attrs(struct msgb *msg, const struct gsm_bts *bts)
+/* Put NM_ATT_SW_CONFIG as per 9.4.61 "SW Configuration" */
+static int add_att_sw_config(struct msgb *msg, const struct gsm_abis_mo *mo)
 {
-	uint16_t total_len = 0;
 	uint8_t *len;
 
 	/* Put NM_ATT_SW_CONFIG as per 9.4.61 "SW Configuration". */
@@ -175,17 +175,39 @@ static inline void add_bts_attrs(struct msgb *msg, const struct gsm_bts *bts)
 	/* We don't know the length yet, so we update it later. */
 	len = msgb_put(msg, 2);
 
-	total_len += abis_nm_put_sw_file(msg, "osmobts", PACKAGE_VERSION, true);
-	total_len += abis_nm_put_sw_file(msg, btsatttr2str(BTS_TYPE_VARIANT),
-					 btsvariant2str(bts->variant), true);
+	switch (mo->obj_class) {
+	case NM_OC_BTS:
+	{
+		const struct gsm_bts *bts = mo->bts;
 
-	if (strlen(bts->sub_model)) {
-		total_len += abis_nm_put_sw_file(msg, btsatttr2str(BTS_SUB_MODEL),
-						 bts->sub_model, true);
+		abis_nm_put_sw_file(msg, "osmobts", PACKAGE_VERSION, true);
+		abis_nm_put_sw_file(msg, btsatttr2str(BTS_TYPE_VARIANT),
+				    btsvariant2str(bts->variant), true);
+		if (strlen(bts->sub_model)) {
+			abis_nm_put_sw_file(msg, btsatttr2str(BTS_SUB_MODEL),
+					    bts->sub_model, true);
+		}
+		break;
+	}
+	case NM_OC_BASEB_TRANSC:
+	{
+		const struct gsm_bts_trx *trx = container_of(mo, struct gsm_bts_trx, bb_transc.mo);
+		const struct phy_instance *pinst = trx->pinst;
+		const char *phy_version;
+
+		phy_version = pinst && strlen(pinst->version) ? pinst->version : "Unknown";
+		abis_nm_put_sw_file(msg, btsatttr2str(TRX_PHY_VERSION), phy_version, true);
+		break;
+	}
+	default:
+		msgb_get(msg, 1 + 2); /* TL16 */
+		return -ENOTSUP;
 	}
 
 	/* Finally, update the length */
-	osmo_store16be(total_len, len);
+	osmo_store16be((uint16_t)(msg->tail - (len + 2)), len);
+
+	return 0;
 }
 
 /* Add BTS features as 3GPP TS 52.021 §9.4.30 Manufacturer Id */
@@ -195,98 +217,249 @@ static inline void add_bts_feat(struct msgb *msg, const struct gsm_bts *bts)
 	msgb_tl16v_put(msg, NM_ATT_MANUF_ID, len, bts->features->data);
 }
 
-static inline void add_trx_attr(struct msgb *msg, const struct gsm_bts_trx *trx)
+/* Add ip.access feature flags for the given MO */
+static int add_att_ipacc_features(struct msgb *msg, const struct gsm_abis_mo *mo)
 {
-	const struct phy_instance *pinst = trx_phy_instance(trx);
-	const char *phy_version;
-	uint16_t total_len;
+	const struct gsm_bts *bts = mo->bts;
+	const struct gsm_bts_trx *trx;
+	uint32_t val;
 	uint8_t *len;
 
-	/* Put NM_ATT_SW_CONFIG as per 9.4.61 "SW Configuration". */
-	msgb_v_put(msg, NM_ATT_SW_CONFIG);
+	msgb_v_put(msg, NM_ATT_IPACC_SUPP_FEATURES);
 
 	/* We don't know the length yet, so we update it later. */
 	len = msgb_put(msg, 2);
 
-	phy_version = pinst && strlen(pinst->version) ? pinst->version : "Unknown";
-	total_len = abis_nm_put_sw_file(msg, btsatttr2str(TRX_PHY_VERSION), phy_version, true);
+	switch (mo->obj_class) {
+	case NM_OC_BTS:
+		msgb_tv16_put(msg, NM_IPAC_EIE_MAX_TA, 1); /* TL16 */
+		msgb_put_u8(msg, (bts->support.max_ta >> 0) & 0xff);
+		break;
+	case NM_OC_RADIO_CARRIER:
+		trx = container_of(mo, struct gsm_bts_trx, mo);
+		msgb_tv16_put(msg, NM_IPAC_EIE_FREQ_BANDS, 1); /* TL16 */
+		msgb_put_u8(msg, (trx->support.freq_bands >> 0) & 0xff);
+		break;
+	case NM_OC_BASEB_TRANSC:
+		trx = container_of(mo, struct gsm_bts_trx, bb_transc.mo);
+		msgb_tv16_put(msg, NM_IPAC_EIE_CIPH_ALGOS, 1); /* TL16 */
+		msgb_put_u8(msg, bts->support.ciphers); /* LSB is A5/1 */
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_CHAN_TYPES, 2); /* TL16 */
+		msgb_put_u8(msg, (trx->support.chan_types >> 0) & 0xff);
+		msgb_put_u8(msg, (trx->support.chan_types >> 8) & 0xff);
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_CHAN_MODES, 3); /* TL16 */
+		msgb_put_u8(msg, (trx->support.chan_modes >>  0) & 0xff);
+		msgb_put_u8(msg, (trx->support.chan_modes >>  8) & 0xff);
+		msgb_put_u8(msg, (trx->support.chan_modes >> 16) & 0xff);
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_RTP_FEATURES, 1); /* TL16 */
+		val = NM_IPAC_F_RTP_FEAT_IR_64k;
+		msgb_put_u8(msg, (val >> 0) & 0xff);
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_RSL_FEATURES, 1); /* TL16 */
+		val = NM_IPAC_F_RSL_FEAT_DYN_PDCH_ACT
+		    | NM_IPAC_F_RSL_FEAT_RTP_PT2;
+		msgb_put_u8(msg, (val >> 0) & 0xff);
+		break;
+	case NM_OC_GPRS_CELL:
+		msgb_tv16_put(msg, NM_IPAC_EIE_GPRS_CODING, 2); /* TL16 */
+		msgb_put_u8(msg, (bts->gprs.cell.support.gprs_codings >> 0) & 0xff);
+		msgb_put_u8(msg, (bts->gprs.cell.support.gprs_codings >> 8) & 0xff);
+		break;
+	default:
+		msgb_get(msg, 1 + 2); /* TL16 */
+		return -ENOTSUP;
+	}
 
 	/* Finally, update the length */
-	osmo_store16be(total_len, len);
-}
-
-/* Handle a list of attributes requested by the BSC, compose
- * TRX-specific Get Attribute Response IE as per 9.4.64. */
-static inline int handle_attrs_trx(struct msgb *out_msg, const struct gsm_bts_trx *trx,
-				   const uint8_t *attr, uint16_t attr_len)
-{
-	uint8_t num_unsupported = 0;
-	uint8_t *buf;
-	int i;
-
-	if (!trx) {
-		LOGP(DOML, LOGL_ERROR, "%s: O&M Get Attributes for unknown TRX\n", gsm_trx_name(trx));
-		return -NM_NACK_TRXNR_UNKN;
-	}
-
-	for (i = 0; i < attr_len; i++) {
-		switch (attr[i]) {
-		case NM_ATT_SW_CONFIG:
-			add_trx_attr(out_msg, trx);
-			break;
-		default:
-			LOGP(DOML, LOGL_ERROR, "%s: O&M Get Attributes [%u], %s is unsupported by TRX\n",
-			     gsm_trx_name(trx), i, get_value_string(abis_nm_att_names, attr[i]));
-			/* Push this tag to the list of unsupported attributes */
-			buf = msgb_push(out_msg, 1);
-			*buf = attr[i];
-			num_unsupported++;
-		}
-	}
-
-	/* Push the amount of unsupported attributes */
-	buf = msgb_push(out_msg, 1);
-	*buf = num_unsupported;
+	osmo_store16be((uint16_t)(msg->tail - (len + 2)), len);
 
 	return 0;
 }
 
-/* Handle a list of attributes requested by the BSC, compose
- * BTS-specific Get Attribute Response IE as per 9.4.64. */
-static inline int handle_attrs_bts(struct msgb *out_msg, const struct gsm_bts *bts,
-				   const uint8_t *attr, uint16_t attr_len)
+/* Add attribute 9.4.8 BCCH ARFCN for BTS class */
+static inline void add_att_bcch_arfcn(struct msgb *msg, const struct gsm_bts *bts)
 {
-	uint8_t num_unsupported = 0;
-	uint8_t *buf;
-	int i;
+	/* type + 16 bit value */
+	msgb_tv16_put(msg, NM_ATT_BCCH_ARFCN, bts->c0->arfcn);
+}
 
-	if (!bts) {
-		LOGP(DOML, LOGL_ERROR, "O&M Get Attributes for unknown BTS\n");
-		return -NM_NACK_BTSNR_UNKN;
+/* Add attribute 9.4.25 Interference Level Boundaries for BTS class */
+static inline void add_att_interf_bound(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit values */
+	msgb_put_u8(msg, NM_ATT_INTERF_BOUND);
+	for (int j = 0; j < ARRAY_SIZE(bts->interference.boundary); j++)
+		msgb_put_u8(msg, abs(bts->interference.boundary[j]));
+}
+
+/* Add attribute 9.4.24 Intave Parameter for BTS class */
+static inline void add_att_intave_param(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_INTAVE_PARAM, bts->interference.intave);
+}
+
+/* Add attribute 9.4.14 Connection Failure Criterion for BTS class */
+static inline void add_att_conn_fail_crit(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + length + values */
+	msgb_tv16_put(msg, NM_ATT_CONN_FAIL_CRIT, 2);
+	msgb_put_u8(msg, 0x01);
+	msgb_put_u8(msg, bts->radio_link_timeout.current);
+}
+
+/* Add attribute 9.4.31 Maximum Timing Advance for BTS class */
+static inline void add_att_max_ta(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_MAX_TA, bts->max_ta);
+}
+
+/* Add attribute 9.4.39 Overload Period for BTS class */
+static inline void add_att_overl_period(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + length + value */
+	msgb_tv16_put(msg, NM_ATT_OVERL_PERIOD, 1);
+	msgb_put_u8(msg, bts->load.overload_period);
+}
+
+/* Add attribute 9.4.12 CCCH Load Threshold for BTS class */
+static inline void add_att_ccch_l_t(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_CCCH_L_T, bts->load.ccch.load_ind_thresh);
+}
+
+/* Add attribute 9.4.11 CCCH Load Indication Period for BTS class */
+static inline void add_att_ccch_l_i_p(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_CCCH_L_I_P, bts->load.ccch.load_ind_period);
+}
+
+/* Add attribute 9.4.44 RACH Busy Threshold for BTS class */
+static inline void add_att_rach_b_thresh(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_RACH_B_THRESH, abs(bts->load.rach.busy_thresh));
+}
+
+/* Add attribute 9.4.45 RACH Load Averaging Slots for BTS class */
+static inline void add_att_ldavg_slots(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 16 bit value */
+	msgb_tv16_put(msg, NM_ATT_LDAVG_SLOTS, bts->load.rach.averaging_slots);
+}
+
+/* Add attribute 9.4.10 BTS Air Timer for BTS class */
+static inline void add_att_bts_air_timer(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_BTS_AIR_TIMER, bts->t3105_ms / 10);
+}
+
+/* Add attribute 9.4.37 NY1 for BTS class */
+static inline void add_att_ny1(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_NY1, bts->ny1);
+}
+
+/* Add attribute 9.4.9 BSIC for BTS class */
+static inline int add_att_bsic(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* BSIC must be configured. */
+	if (!bts->bsic_configured)
+		return -EINVAL;
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_BSIC, bts->bsic);
+	return 0;
+}
+
+/* Add attribute 9.4.20 GSM Time for BTS class */
+static inline void add_att_gsm_time(struct msgb *msg, const struct gsm_bts *bts)
+{
+	/* type + 16 bit value */
+	msgb_tv16_put(msg, NM_ATT_GSM_TIME, bts->gsm_time.fn % GSM_RFN_MODULUS);
+}
+
+/* Add attribute 9.4.47 RF Max Power Reduction for radio carrier class */
+static inline void add_att_rf_maxpowr_r(struct msgb *msg, const struct gsm_bts_trx *trx)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_RF_MAXPOWR_R, trx->max_power_red / 2);
+}
+
+/* Add attribute 9.4.5 ARFCN List for radio carrier class */
+static inline void add_att_arfcn_list(struct msgb *msg, const struct gsm_bts_trx *trx)
+{
+#if 0
+	/* type + length + values */
+	msgb_tv16_put(msg, NM_ATT_ARFCN_LIST, trx->arfcn_num * 2);
+	for (int j = 0; j < trx->arfcn_num; j++)
+		msgb_put_u16(msg, trx->arfcn_list[j]);
+#else
+	/* type + length + values */
+	msgb_tv16_put(msg, NM_ATT_ARFCN_LIST, 2);
+	msgb_put_u16(msg, trx->arfcn);
+#endif
+}
+
+/* Add attribute 9.4.5 ARFCN List for channel class */
+static inline void add_att_arfcn_list_ts(struct msgb *msg, const struct gsm_bts_trx_ts *ts)
+{
+	if (ts->hopping.enabled) {
+		/* type + length + values */
+		msgb_tv16_put(msg, NM_ATT_ARFCN_LIST, ts->hopping.arfcn_num * 2);
+		for (int j = 0; j < ts->hopping.arfcn_num; j++)
+			msgb_put_u16(msg, ts->hopping.arfcn_list[j]);
+	} else {
+		/* type + length + values */
+		msgb_tv16_put(msg, NM_ATT_ARFCN_LIST, 2);
+		msgb_put_u16(msg, ts->trx->arfcn);
 	}
+}
 
-	for (i = 0; i < attr_len; i++) {
-		switch (attr[i]) {
-		case NM_ATT_SW_CONFIG:
-			add_bts_attrs(out_msg, bts);
-			break;
-		case NM_ATT_MANUF_ID:
-			add_bts_feat(out_msg, bts);
-			break;
-		default:
-			LOGP(DOML, LOGL_ERROR, "O&M Get Attributes [%u], %s is unsupported by BTS\n", i,
-			     get_value_string(abis_nm_att_names, attr[i]));
-			/* Push this tag to the list of unsupported attributes */
-			buf = msgb_push(out_msg, 1);
-			*buf = attr[i];
-			num_unsupported++;
-		}
-	}
+/* Add attribute 9.4.13 Channel Combination for channel class */
+static inline int add_att_chan_comb(struct msgb *msg, const struct gsm_bts_trx_ts *ts)
+{
+	int comb = abis_nm_chcomb4pchan(ts->pchan);
 
-	/* Push the amount of unsupported attributes */
-	buf = msgb_push(out_msg, 1);
-	*buf = num_unsupported;
+	/* If current channel combination is not yet set, 0xff is returned. */
+	if (comb < 0 || comb == 0xff)
+		return -EINVAL;
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_CHAN_COMB, comb);
+	return 0;
+}
 
+/* Add attribute 9.4.60 TSC for channel class */
+static inline void add_att_tsc(struct msgb *msg, const struct gsm_bts_trx_ts *ts)
+{
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_TSC, ts->tsc);
+}
+
+/* Add attribute 9.4.60 HSN for channel class */
+static inline int add_att_hsn(struct msgb *msg, const struct gsm_bts_trx_ts *ts)
+{
+	if (!ts->hopping.enabled)
+		return -EINVAL;
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_HSN, ts->hopping.hsn);
+	return 0;
+}
+
+/* Add attribute 9.4.21 MAIO for channel class */
+static inline int add_att_maio(struct msgb *msg, const struct gsm_bts_trx_ts *ts)
+{
+	if (!ts->hopping.enabled)
+		return -EINVAL;
+	/* type + 8 bit value */
+	msgb_tv_put(msg, NM_ATT_MAIO, ts->hopping.maio);
 	return 0;
 }
 
@@ -295,32 +468,177 @@ static int oml_tx_attr_resp(const struct gsm_abis_mo *mo,
 			    const uint8_t *attr, uint16_t attr_len)
 {
 	struct msgb *nmsg = oml_msgb_alloc();
-	const char *mo_name = gsm_abis_mo_name(mo);
+	unsigned int num_unsupported = 0;
+	struct gsm_bts_trx *trx = NULL;
+	struct gsm_bts_trx_ts *ts = NULL;
 	int rc;
 
 	if (!nmsg)
 		return -NM_NACK_CANT_PERFORM;
 
+	/* Set TRX, if object class is Radio Carrier, Baseband Transceiver or Channel. */
 	switch (mo->obj_class) {
-	case NM_OC_BTS:
-		rc = handle_attrs_bts(nmsg, mo->bts, attr, attr_len);
-		break;
+	case NM_OC_RADIO_CARRIER:
 	case NM_OC_BASEB_TRANSC:
-		rc = handle_attrs_trx(nmsg, gsm_bts_trx_num(mo->bts, mo->obj_inst.trx_nr), attr, attr_len);
+	case NM_OC_CHANNEL:
+		trx = gsm_bts_trx_num(mo->bts, mo->obj_inst.trx_nr);
 		break;
-	default:
-		LOGP(DOML, LOGL_ERROR, "%s: Unsupported MO class in Get Attribute Response\n",
-		     mo_name);
-		rc = -NM_NACK_OBJCLASS_NOTSUPP;
 	}
 
-	if (rc < 0) {
-		LOGP(DOML, LOGL_ERROR, "%s: Tx Get Attribute Response FAILED with rc=%d\n",
-		     mo_name, rc);
-		msgb_free(nmsg);
-		return rc;
+	/* Set TS, if object class is Channel. */
+	if (mo->obj_class == NM_OC_CHANNEL && trx)
+		ts = &trx->ts[mo->obj_inst.ts_nr];
+
+	for (unsigned int i = 0; i < attr_len; i++) {
+		switch (attr[i]) {
+		case NM_ATT_OPER_STATE:
+			msgb_tv16_put(nmsg, attr[i], 1);
+			msgb_put_u8(nmsg, mo->nm_state.operational);
+			break;
+		case NM_ATT_ADM_STATE:
+			msgb_tv16_put(nmsg, attr[i], 1);
+			msgb_put_u8(nmsg, mo->nm_state.administrative);
+			break;
+		case NM_ATT_AVAIL_STATUS:
+			msgb_tv16_put(nmsg, attr[i], 1);
+			msgb_put_u8(nmsg, mo->nm_state.availability);
+			break;
+		case NM_ATT_SW_CONFIG:
+			if (add_att_sw_config(nmsg, mo) != 0)
+				goto unsupported;
+			break;
+		case NM_ATT_MANUF_ID:
+			if (mo->obj_class == NM_OC_BTS)
+				add_bts_feat(nmsg, mo->bts);
+			else
+				goto unsupported;
+			break;
+		case NM_ATT_IPACC_SUPP_FEATURES:
+			if (add_att_ipacc_features(nmsg, mo) != 0)
+				goto unsupported;
+			break;
+		case NM_ATT_BCCH_ARFCN:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_bcch_arfcn(nmsg, mo->bts);
+			break;
+		case NM_ATT_INTERF_BOUND:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_interf_bound(nmsg, mo->bts);
+			break;
+		case NM_ATT_INTAVE_PARAM:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_intave_param(nmsg, mo->bts);
+			break;
+		case NM_ATT_CONN_FAIL_CRIT:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_conn_fail_crit(nmsg, mo->bts);
+			break;
+		case NM_ATT_MAX_TA:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_max_ta(nmsg, mo->bts);
+			break;
+		case NM_ATT_OVERL_PERIOD:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_overl_period(nmsg, mo->bts);
+			break;
+		case NM_ATT_CCCH_L_T:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_ccch_l_t(nmsg, mo->bts);
+			break;
+		case NM_ATT_CCCH_L_I_P:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_ccch_l_i_p(nmsg, mo->bts);
+			break;
+		case NM_ATT_RACH_B_THRESH:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_rach_b_thresh(nmsg, mo->bts);
+			break;
+		case NM_ATT_LDAVG_SLOTS:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_ldavg_slots(nmsg, mo->bts);
+			break;
+		case NM_ATT_BTS_AIR_TIMER:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_bts_air_timer(nmsg, mo->bts);
+			break;
+		case NM_ATT_NY1:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_ny1(nmsg, mo->bts);
+			break;
+		case NM_ATT_BSIC:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			if (add_att_bsic(nmsg, mo->bts) != 0)
+				goto unsupported;
+			break;
+		case NM_ATT_GSM_TIME:
+			if (mo->obj_class != NM_OC_BTS)
+				goto unsupported;
+			add_att_gsm_time(nmsg, mo->bts);
+			break;
+		case NM_ATT_RF_MAXPOWR_R:
+			if (mo->obj_class != NM_OC_RADIO_CARRIER || !trx)
+				goto unsupported;
+			add_att_rf_maxpowr_r(nmsg, trx);
+			break;
+		case NM_ATT_ARFCN_LIST:
+			if (mo->obj_class == NM_OC_RADIO_CARRIER && trx) {
+				add_att_arfcn_list(nmsg, trx);
+				break;
+			}
+			if (mo->obj_class == NM_OC_CHANNEL && ts) {
+				add_att_arfcn_list_ts(nmsg, ts);
+				break;
+			}
+			goto unsupported;
+		case NM_ATT_CHAN_COMB:
+			if (mo->obj_class != NM_OC_CHANNEL || !ts)
+				goto unsupported;
+			if (add_att_chan_comb(nmsg, ts) != 0)
+				goto unsupported;
+			break;
+		case NM_ATT_TSC:
+			if (mo->obj_class != NM_OC_CHANNEL || !ts)
+				goto unsupported;
+			add_att_tsc(nmsg, ts);
+			break;
+		case NM_ATT_HSN:
+			if (mo->obj_class != NM_OC_CHANNEL || !ts)
+				goto unsupported;
+			if (add_att_hsn(nmsg, ts) != 0)
+				goto unsupported;
+			break;
+		case NM_ATT_MAIO:
+			if (mo->obj_class != NM_OC_CHANNEL || !ts)
+				goto unsupported;
+			if (add_att_maio(nmsg, ts) != 0)
+				goto unsupported;
+			break;
+		default:
+unsupported:
+			LOGP(DOML, LOGL_ERROR, "%s: O&M Get Attributes [%u], %s is unsupported\n",
+			     gsm_abis_mo_name(mo), i, get_value_string(abis_nm_att_names, attr[i]));
+			/* Push this tag to the list of unsupported attributes */
+			msgb_push_u8(nmsg, attr[i]);
+			num_unsupported++;
+		}
 	}
 
+	/* Push the amount of unsupported attributes */
+	msgb_push_u8(nmsg, num_unsupported);
+
 	/* Push Get Attribute Response Info TL (actually TV where V is L) */
 	msgb_tv16_push(nmsg, NM_ATT_GET_ARI, msgb_length(nmsg));
 
@@ -515,16 +833,21 @@ int oml_mo_tx_sw_act_rep(const struct gsm_abis_mo *mo)
 	return oml_mo_send_msg(mo, nmsg, NM_MT_SW_ACTIVATED_REP);
 }
 
-/* The defaults below correspond to various sources/recommendations that could be found online.
- * The BSC should override this via OML anyway. */
-const unsigned int oml_default_t200_ms[7] = {
-        [T200_SDCCH]		= 1000,
-        [T200_FACCH_F]		= 1000,
-        [T200_FACCH_H]		= 1000,
-        [T200_SACCH_TCH_SAPI0]	= 2000,
-        [T200_SACCH_SDCCH]	= 2000,
-        [T200_SDCCH_SAPI3]	= 1000,
-        [T200_SACCH_TCH_SAPI3]	= 2000,
+/* The defaults below correspond to the number of frames until a response from the MS is expected.
+ * It defines the FN distance between the frame number when a message is sent (first frame) and when the response is
+ * received (first frame). On SACCH the duration is two frames, because SAPI0 and SAPI3 are are transmitted in
+ * alternating order. On DCCH with SAPI3 the duration is two seconds, because SAPI0 has priority over SAPI3.
+ *
+ * See Table 8 if 3GPP TS 44.006. Note that the table only shows the FN distance between frames.
+ */
+const uint32_t oml_default_t200_fn[7] = {
+	[T200_SDCCH]		= 4+32,
+	[T200_FACCH_F]		= 8+9,
+	[T200_FACCH_H]		= 6+10,
+	[T200_SACCH_TCH_SAPI0]	= 79+25+104,
+	[T200_SACCH_SDCCH]	= 4+32+51,
+	[T200_SDCCH_SAPI3]	= 4+32+408, /* two seconds */
+	[T200_SACCH_TCH_SAPI3]	= 79+25+104,
 };
 
 /* 3GPP TS 52.021 §8.11.1 Get Attributes has been received */
@@ -667,25 +990,26 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 	}
 
 	/* 9.4.53 T200 */
-	if (TLVP_PRES_LEN(&tp, NM_ATT_T200, ARRAY_SIZE(bts->t200_ms))) {
-		payload = TLVP_VAL(&tp, NM_ATT_T200);
-		for (i = 0; i < ARRAY_SIZE(bts->t200_ms); i++) {
-			uint32_t t200_ms = payload[i] * abis_nm_t200_ms[i];
+	if (TLVP_PRES_LEN(&tp, NM_ATT_T200, ARRAY_SIZE(bts->t200_fn))) {
+		/* The OML message NM_ATT_T200 is ignored, because T200 timeouts are set to
+		 * the minimal response time. Longer timeouts would cause lower throughput
+		 * in case of lost frames. Shorter timeouts would cause LAPDm to fail. */
+		DEBUGPFOH(DOML, foh, "Ignoring T200 BTS attribute.\n");
 #if 0
-			bts->t200_ms[i] = t200_ms;
-			DEBUGPFOH(DOML, foh, "T200[%u]: OML=%u, mult=%u => %u ms\n",
+		payload = TLVP_VAL(&tp, NM_ATT_T200);
+		for (i = 0; i < ARRAY_SIZE(bts->t200_fn); i++) {
+			uint32_t t200_ms = payload[i] * abis_nm_t200_ms[i];
+			uint32_t t200_fn = t200_ms * 1000 + (GSM_TDMA_FN_DURATION_uS - 1) / GSM_TDMA_FN_DURATION_uS;
+			/* Values must not be less than absolute minimum. */
+			if (oml_default_t200_fn[i] <= t200_fn)
+				bts->t200_fn[i] = t200_fn;
+			else
+				bts->t200_fn[i] = oml_default_t200_fn[i];
+			DEBUGPFOH(DOML, foh, "T200[%u]: OML=%u, mult=%u => %u ms -> %u fn\n",
 				  i, payload[i], abis_nm_t200_ms[i],
-				  bts->t200_ms[i]);
-#else
-                        /* we'd rather use the 1s/2s (long) defaults by
-                         * libosmocore, as we appear to have some bug(s)
-                         * related to handling T200 expiration in
-                         * libosmogsm lapd(m) code? */
-                        LOGPFOH(DOML, LOGL_NOTICE, foh, "Ignoring T200[%u] (%u ms) "
-                                "as sent by BSC due to suspected LAPDm bug!\n",
-                                i, t200_ms);
-#endif
+				  t200_ms, bts->t200_fn[i]);
 		}
+#endif
 	}
 
 	/* 9.4.31 Maximum Timing Advance */
@@ -1534,6 +1858,7 @@ static int rx_oml_ipa_rsl_connect(struct gsm_bts *bts, struct msgb *msg,
 	struct e1inp_sign_link *oml_link = bts->oml_link;
 	const struct abis_om_fom_hdr *foh = msgb_l3(msg);
 	struct gsm_bts_trx *trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
+	struct gsm_bts_bb_trx *bb_transc;
 	const char *trx_name;
 	struct in_addr in;
 	uint16_t port = IPA_TCP_PORT_RSL;
@@ -1547,6 +1872,7 @@ static int rx_oml_ipa_rsl_connect(struct gsm_bts *bts, struct msgb *msg,
 
 	if (TLVP_PRESENT(tp, NM_ATT_IPACC_DST_IP_PORT))
 		port = ntohs(tlvp_val16_unal(tp, NM_ATT_IPACC_DST_IP_PORT));
+
 	if (TLVP_PRESENT(tp, NM_ATT_IPACC_STREAM_ID))
 		stream_id = *TLVP_VAL(tp, NM_ATT_IPACC_STREAM_ID);
 
@@ -1557,22 +1883,16 @@ static int rx_oml_ipa_rsl_connect(struct gsm_bts *bts, struct msgb *msg,
 		goto tx_ack_nack;
 	}
 
+	bb_transc = &trx->bb_transc;
+	osmo_sockaddr_str_from_in_addr(&bb_transc->rsl.rem_addrstr, &in, port);
+	bb_transc->rsl.tei = stream_id;
+
 	trx_name = gsm_trx_name(trx);
-
 	LOGP(DOML, LOGL_INFO, "%s: Rx IPA RSL CONNECT IP=%s PORT=%u STREAM=0x%02x\n",
-	     trx_name, inet_ntoa(in), port, stream_id);
+	     trx_name, bb_transc->rsl.rem_addrstr.ip, bb_transc->rsl.rem_addrstr.port,
+	     bb_transc->rsl.tei);
 
-	if (bts->variant == BTS_OSMO_OMLDUMMY) {
-		rc = 0;
-		LOGP(DOML, LOGL_NOTICE, "%s: Not connecting RSL in OML-DUMMY!\n", trx_name);
-	} else {
-		trx->rsl_tei = stream_id;
-		rc = e1inp_ipa_bts_rsl_connect_n(oml_link->ts->line, inet_ntoa(in), port, trx->nr);
-		if (rc < 0) {
-			LOGP(DOML, LOGL_NOTICE, "%s: Error connecting IPA RSL: %d\n", trx_name, rc);
-			rc = NM_NACK_CANT_PERFORM;
-		}
-	}
+	rc = 0;
 
 tx_ack_nack:
 	/* The ACK/NACK is expected to contain all IEs */
