@@ -11,7 +11,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -1172,8 +1172,8 @@ int lchan_activate(struct gsm_lchan *lchan)
 		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Trying to activate lchan, but commands in queue\n");
 
 	/* For handover, always start the main channel immediately. lchan->want_dl_sacch_active indicates whether dl
-	 * SACCH should be activated. Also, for HO and VGCS listener/talker detection, start the RACH SAPI. */
-	if (lchan->ho.active == HANDOVER_ENABLED || rsl_chan_rt_is_asci(lchan->rsl_chan_rt))
+	 * SACCH should be activated. */
+	if (lchan->ho.active == HANDOVER_ENABLED)
 		enqueue_sapi_act_cmd(lchan, GsmL1_Sapi_Rach, GsmL1_Dir_RxUplink);
 
 	for (i = 0; i < s4l->num_sapis; i++) {
@@ -1470,6 +1470,16 @@ int l1if_set_ciphering(struct femtol1_hdl *fl1h,
 	return 0;
 }
 
+int l1if_set_ul_acc(struct gsm_lchan *lchan, bool active)
+{
+	if (active)
+		enqueue_sapi_act_cmd(lchan, GsmL1_Sapi_Rach, GsmL1_Dir_RxUplink);
+	else
+		check_sapi_release(lchan, GsmL1_Sapi_Rach, GsmL1_Dir_RxUplink);
+
+	return 0;
+}
+
 int bts_model_adjst_ms_pwr(struct gsm_lchan *lchan)
 {
 	if (lchan->state != LCHAN_S_ACTIVE)
